import pusion as p
from pusion.auto.detector import *
from pusion.auto.generic_combiner import GenericCombiner
from pusion.evaluation.evaluation_metrics import *
from pusion.util.generator import *


class AutoCombiner(GenericCombiner):
    """
    The `AutoCombiner` allows for automatic decision fusion using all methods provided by the framework, which are
    applicable to the given problem. The key feature of this combiner is the transparency in terms of it's outer
    behaviour. Based on the usage (i.e. method calls) and the automatically detected configuration,
    the `AutoCombiner` preselects all compatible methods from `pusion.core`. The main purpose is to retrieve fusion
    results obtained by the methods with the best performance without further user interaction.
    """

    _SUPPORTED_PAC = [
        (Problem.GENERIC, AssignmentType.GENERIC, CoverageType.GENERIC),
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT)
    ]

    def __init__(self):
        super().__init__()
        self.selected_combiner = None
        self.validation_size = 0.5
        self.eval_metric = p.PerformanceMetric.ACCURACY # default evaluation metric for the AutoFusion approach

    def train(self, decision_tensor, true_assignments, **kwargs):
        """
        Train the AutoCombiner (AC) model. This method detects the configuration based on the ``decision_tensor`` and
        trains all trainable combiners that are applicable to this configuration.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.

                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """

        # Split into train and validation data.
        if 'test_data' in kwargs.keys():
            dt_train = decision_tensor
            ta_train = true_assignments
            dt_valid = kwargs['test_data']['dt_valid']
            ta_valid = kwargs['test_data']['ta_valid']
        else:
            dt_train, ta_train, dt_valid, ta_valid = split_into_train_and_validation_data(decision_tensor,
                                                                                          true_assignments,
                                                                                          self.validation_size)

        # set specified evaluation metric
        if 'eval_metric' in kwargs.keys():
            self.eval_metric = kwargs['eval_metric']

        # Encapsulated training phase.
        super().train(dt_train, ta_train)
        # Encapsulated evaluation phase.
        super().combine(dt_valid)
        # Todo save the decision fusion evaluation report within the AutoFusion procedure to be able to print the
        #  overview of the performance of the tested decision fusion methods
        performance_per_combiner = np.zeros(len(self.combiners))
        for i in range(len(self.combiners)):
            comb_res = self.multi_combiner_decision_tensor[i]
            # TODO add support for a combined evaluation metric a la Alejandro Villanuevas Paper in the 3d space
            #  --> add a weighted distance-based selection approach to select the best performing decision fusion algorithm, e. g. acc, fdr and prbabilistic prediction quality
            #  also further approaches, e. g. significance tests, cutomized cost functions with weightings ...
            #performance_per_combiner[i] = accuracy(ta_valid, comb_res)
            performance_per_combiner[i] = self.eval_metric(ta_valid, comb_res)
        self.selected_combiner = self.combiners[performance_per_combiner.argmax()]
        # Clear temporarily obtained fusion results.
        self.multi_combiner_decision_tensor = []
        # Here, the AutoCombiner could be trained on the whole dataset again.
        # super().train(decision_tensor, true_assignments)

    def combine(self, decision_tensor):
        """
        Combine decision outputs using the AutoCombiner (AC) model. Both continuous and crisp classification outputs are
        supported. This procedure involves selecting the best method regarding its classification performance in case
        of a trained AC.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.

                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of crisp or continuous class assignments which represents fused decisions.
                Axis 0 represents samples and axis 1 the class labels which are aligned with axis 2 in
                ``decision_tensor`` input tensor.
        """
        #super().combine(decision_tensor)  # Left for insights into performances of preselected combiners.

        if self.selected_combiner is not None:
            return self.selected_combiner.combine(decision_tensor)
        else:
            raise TypeError("No selection performed. Use train() before combining to obtain an automatic selection.")

    def set_validation_size(self, validation_size):
        """
        Set the validation size, based on which the training data is split and the best combiner is selected.

        :param validation_size: A `float` between `0` and `1.0`. Ratio of the validation data set.
        """
        self.validation_size = validation_size

    def get_selected_combiner(self):
        """
        :return: The method selected by the `AutoCombiner`.
        """
        return self.selected_combiner

    def get_eval_metric(self):
        """
        :return: The metric used for the selection of the best performing combiner.
        """
        return self.eval_metric



import numpy as np
#import torch
import torchmetrics
from sklearn.metrics import *

from pusion.auto.detector import determine_problem
from pusion.util.constants import Problem
from pusion.util.transformer import multiclass_assignments_to_labels, multilabel_to_multiclass_assignments


def multi_label_brier_score_micro(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Calculate the brier score for multi-label problems according to Brier 1950
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The micro brier score.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    y_pred_flatten = y_pred.flatten()
    y_target_flatten = y_true.flatten()
    result = np.sum((y_pred_flatten - y_target_flatten) ** 2) / y_pred_flatten.shape[0]
    result = result / y_pred_flatten.shape[0]

    return result


def multi_label_brier_score(y_true: np.ndarray, y_pred: np.ndarray):  # -> float: #TODO
    """
    Calculate the brier score for multiclass problems according to Brier 1950
    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The brier score.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    result = np.mean(np.sum((y_pred - y_true) ** 2, axis=1))

    return result


def multiclass_brier_score(y_true: np.ndarray, y_pred: np.ndarray):  # -> float:
    """
    Calculate the brier score for multi-label problems according to Brier 1950
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The brier score.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    result = np.mean(np.sum((y_pred - y_true) ** 2, axis=1))

    return result


def far(y_true: np.ndarray, y_pred: np.ndarray, pos_normal_class: int = 0) -> float:
    """
    Calculate the false alarm rate for multiclass and multi-label problems.
    FAR = (number of normal class samples incorrectly classified)/(number of all normal class samples) * 100
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :param pos_normal_class: the position of the 'normal class' in :param y_true and :param y_pred. Default is `0`
    :return: The false alarm rate.
    """
    # False Alarm rate
    # FAR = (number of normal class samples incorrectly classified)/(number of all normal class samples) * 100

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    y_true_normal = y_true[:, pos_normal_class]
    y_pred_normal = y_pred[:, pos_normal_class]

    temp_vec = y_true_normal * y_pred_normal
    num_all_normal_class_samples = np.sum(y_true_normal)
    num_normal_class_samples_incorrectly_classified = num_all_normal_class_samples - np.sum(temp_vec)

    # for testing
    # yr = yr = np.array([y_true_normal, y_pred_normal])
    # a = np.unique(yr, axis=1, return_counts=True)
    far = (num_normal_class_samples_incorrectly_classified / num_all_normal_class_samples)
    return far


def multiclass_fdr(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    fault detection rate = (# correctly classified faulty samples) / (# all faulty samples) * 100
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The fault detection rate.
    """
    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    pos_normal_class = 0
    faulty_samples_indices = np.where(y_true[:, pos_normal_class] == 0)[0]
    faulty_samples = y_true[faulty_samples_indices, :]
    pred_samples_at_faulty_indices = y_pred[faulty_samples_indices, :]

    total_num_of_faulty_samples = len(faulty_samples_indices)

    a = faulty_samples != pred_samples_at_faulty_indices
    b = np.sum(a, axis=1)
    uniques, counts = np.unique(b, return_counts=True)
    numbers_of_preds = dict(zip(uniques, counts))

    num_of_correctly_classified_faulty_samples = numbers_of_preds[0]

    fdr = (num_of_correctly_classified_faulty_samples / total_num_of_faulty_samples)
    return fdr


def multilabel_subset_fdr(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    fault detection rate = (# correctly classified faulty samples) / (# all faulty samples) * 100
    In multilabel classification, the function considers the faulty subset, i. e., if the entire set
    of predicted faulty labels for a sample strictly match with the true set of faulty labels.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The fault detection rate.
    """
    # TODO(old): counting: 'subset' or 'minor'. The way how the faulty subsets should be counted. TODO correct?

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    pos_normal_class = 0
    faulty_samples_indices = np.where(y_true[:, pos_normal_class] == 0)[0]
    faulty_samples = y_true[faulty_samples_indices, :]
    pred_samples_at_faulty_indices = y_pred[faulty_samples_indices, :]

    # check faulty samples
    fs1 = faulty_samples[:, 0:pos_normal_class]
    fs2 = faulty_samples[:, pos_normal_class + 1:]
    fs = np.concatenate([fs1, fs2], axis=1)

    # make sure that no sample with ['normal', 0, 0, 0, ..., 0] is contained
    # --> depends on the data set if label 'normal' == [0, 0, ..., 0]
    # --> we assume that label 'normal' == [1, 0, 0, ..., 0]
    fs_sum = np.sum(fs, axis=1)
    fs_indices = np.where(fs_sum > 0)
    fs = fs[fs_indices[0], :]

    if fs.shape[0] != faulty_samples.shape[0]:
        print("Not same length!")

    # fs = faulty_samples #######
    total_num_of_faulty_samples = fs.shape[0]

    fs_pred = pred_samples_at_faulty_indices[fs_indices[0], :]
    fpreds1 = fs_pred[:, 0:pos_normal_class]
    fpreds2 = fs_pred[:, pos_normal_class + 1:]
    fpreds = np.concatenate([fpreds1, fpreds2], axis=1)

    # fpreds = pred_samples_at_faulty_indices ##########
    a = fs != fpreds
    b = np.sum(a, axis=1)
    uniques, counts = np.unique(b, return_counts=True)
    numbers_of_preds = dict(zip(uniques, counts))

    num_of_correctly_classified_faulty_samples = numbers_of_preds[0]
    fdr = (num_of_correctly_classified_faulty_samples / total_num_of_faulty_samples)
    return fdr


def multilabel_minor_fdr(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    fault detection rate = (# correctly classified faulty samples) / (# all faulty samples) * 100
    In multilabel classification, the function considers the faulty subset, i. e., if the entire set
    of predicted faulty labels for a sample strictly match with the true set of faulty labels.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The fault detection rate.
    """
    # TODO(old): counting: 'subset' or 'minor'. The way how the faulty subsets should be counted. TODO correct?

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    pos_normal_class = 0
    faulty_samples_indices = np.where(y_true[:, pos_normal_class] == 0)[0]
    faulty_samples = y_true[faulty_samples_indices, :]
    pred_samples_at_faulty_indices = y_pred[faulty_samples_indices, :]

    # check faulty samples
    fs1 = faulty_samples[:, 0:pos_normal_class]
    fs2 = faulty_samples[:, pos_normal_class + 1:]
    fs = np.concatenate([fs1, fs2], axis=1)

    fs_preds1 = pred_samples_at_faulty_indices[:, 0:pos_normal_class]
    fs_preds2 = pred_samples_at_faulty_indices[:, pos_normal_class + 1:]
    fs_preds = np.concatenate([fs_preds1, fs_preds2], axis=1)

    num_faulty_labels = 0
    num_correctly_classified_faulty_labels = 0
    for col in range(fs.shape[1]):
        fs_col = fs[:, col]
        fs_preds_col = fs_preds[:, col]

        num_faulty_labels = num_faulty_labels + np.sum(fs_col, axis=0)
        temp_mult = fs_col * fs_preds_col
        num_correctly_classified_faulty_labels = num_correctly_classified_faulty_labels + np.sum(temp_mult,
                                                                                                 axis=0)

    fdr = (num_correctly_classified_faulty_labels / num_faulty_labels)
    return fdr


def multiclass_weighted_precision(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Calculate the precision for a multiclass problem with a `weighted` average.
    :param y_true: `numpy.array` of shape `(n_samples,)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)`. Predicted labels or class assignments.
    :return: The precision score.
    """

    if not (y_true.ndim == 1 and y_pred.ndim == 1 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 1D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return precision_score(y_true, y_pred, average='weighted')


def multi_label_weighted_precision(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Calculate the precision for a multi-label problem with a `weighted` average.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The precision score.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return precision_score(y_true, y_pred, average='weighted')


def multiclass_class_wise_precision(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    """
    Calculate the precision for a multiclass problem with average `None`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The precision score.
    """
    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    y_true = multiclass_assignments_to_labels(y_true)
    y_pred = multiclass_assignments_to_labels(y_pred)

    return precision_score(y_true, y_pred, average=None)


def multi_label_class_wise_precision(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    """
    Calculate the precision for a multi-label problem with average `None`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The precision score.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return precision_score(y_true, y_pred, average=None)


def multiclass_recall(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Calculate the recall for a multiclass problem with average `weighted`.
    :param y_true: `numpy.array` of shape `(n_samples,)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)`. Predicted labels or class assignments.
    :return: The recall score.
    """

    if not (y_true.ndim == 1 and y_pred.ndim == 1 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 1D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return recall_score(y_true, y_pred,
                        average='weighted')  # Weighted recall is equal to accuracy. Cf. sk learn doc


def multi_label_recall(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Calculate the recall for a multi-label problem with average `weighted`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The recall score.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return recall_score(y_true, y_pred, average='weighted')  # Weighted recall is equal to accuracy. Cf. sk learn doc


def multiclass_class_wise_recall(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    """
    Calculate the recall for a multiclass problem with average `None`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: Sequence of recall scores (for each class).
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    y_true = multiclass_assignments_to_labels(y_true)
    y_pred = multiclass_assignments_to_labels(y_pred)

    return recall_score(y_true, y_pred, average=None)


def multi_label_class_wise_recall(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    """
    Calculate the recall for a multi-label problem with average `None`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: Sequence of recall scores (for each class).
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return recall_score(y_true, y_pred, average=None)

# TODO float
def multiclass_weighted_scikit_auc_roc_score(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Compute the scikit auc roc score for a multi-label problem with average `weighted`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The auc roc score.
    """

    # TODO: quick fix done here.
    # if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0] and np.issubdtype(y_pred.dtype, np.floating)):
    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0]):
        raise ValueError(
            "y_true and y_pred needs to be 2D with same number of samples and dtype float, "
            "got {} and {} with dtype {} instead.".format(y_true.shape, y_pred.shape, y_pred.dtype)
        )

    # y_true = multiclass_assignments_to_labels(y_true)

    return roc_auc_score(y_true, y_pred, average='weighted', multi_class='ovr')

# TODO float
def multi_label_weighted_pytorch_auc_roc_score(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Compute the pytorch auc roc score for a multi-label problem with average `weighted`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The auc roc score.
    """
    # pytorch auc roc metric
    # TODO: quick fix done here.
    # if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape and np.issubdtype(y_pred.dtype, np.floating)):
    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, y_pred needs to have dtype float, "
            "got {} and {} with dtype {} instead.".format(y_true.shape, y_pred.shape, y_pred.dtype)
        )

    y_pred = y_pred.astype(np.float)
    num_classes = y_true.shape[1]
    y_pred_torch = torch.from_numpy(y_pred)
    y_true_torch = torch.from_numpy(y_true)
    return torchmetrics.functional.auroc(y_pred_torch, y_true_torch, num_classes=num_classes,
                                         average='weighted').numpy().item()

# TODO float
def multi_label_pytorch_auc_roc_score(y_true: np.ndarray, y_pred: np.ndarray): # -> list[float]:
    """
    Compute the pytorch auc roc score for a multi-label problem with average `None`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The auc roc score.
    """
    # pytorch auc roc metric
    # TODO: quick fix done here.
    # if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape and np.issubdtype(y_pred.dtype, np.floating)):
    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, y_pred needs to have dtype float, "
            "got {} and {} with dtype {} instead.".format(y_true.shape, y_pred.shape, y_pred.dtype)
        )
    y_pred = y_pred.astype(np.float)
    num_classes = y_true.shape[1]
    y_pred_torch = torch.from_numpy(y_pred)
    y_true_torch = torch.from_numpy(y_true)
    pytorch_auc_roc_class_wise = torchmetrics.functional.auroc(y_pred_torch, y_true_torch, num_classes=num_classes,
                                                               average=None)
    return [val.numpy().item() for val in pytorch_auc_roc_class_wise]

# TODO float
def multiclass_class_wise_avg_precision(y_true: np.ndarray, y_pred: np.ndarray): # -> list[float]:
    """
    Compute the class wise precision for a multiclass problem with average `None`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The precision score.
    """

    # TODO: quick fix done here.
    # if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0] and np.issubdtype(y_pred.dtype, np.floating)):
    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0]):
        raise ValueError(
            "y_true and y_pred need to be 2D with the same number of samples and dtype float, "
            "got {} and {} with dtype {} instead.".format(y_true.shape, y_pred.shape, y_pred.dtype)
        )

    num_classes = y_true.shape[1]
    y_true = multiclass_assignments_to_labels(y_true)

    y_pred_torch = torch.from_numpy(y_pred)
    y_true_torch = torch.from_numpy(y_true)
    pytorch_average_precision_score = torchmetrics.functional.average_precision(y_pred_torch,
                                                                                y_true_torch,
                                                                                num_classes=num_classes,
                                                                                average=None)
    pytorch_average_precision_score = [val.numpy() for val in pytorch_average_precision_score]
    return [npa.item() for npa in pytorch_average_precision_score]

# TODO float
def multiclass_weighted_avg_precision(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Compute the precision for a multiclass problem with average `weighted`.
    :param y_true: `numpy.array` of shape `(n_samples,)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The precision score.
    """

    if not (y_true.ndim == 1 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0] and np.issubdtype(y_pred.dtype, np.floating)):
        raise ValueError(
            "y_true needs to be 1D and y_pred needs to be 2D with same number of samples and dtype float, "
            "got {} and {} with dtype {} instead.".format(y_true.shape, y_pred.shape, y_pred.dtype)
        )

    num_classes = y_true.shape[1]
    y_pred_torch = torch.from_numpy(y_pred)
    y_true_torch = torch.from_numpy(y_true)
    pytorch_average_precision_score_weighted = torchmetrics.functional.average_precision(y_pred_torch,
                                                                                         y_true_torch,
                                                                                         num_classes=num_classes,
                                                                                         average='weighted')
    return pytorch_average_precision_score_weighted.numpy().item()


# TODO float
def multiclass_auc_precision_recall_curve(y_true: np.ndarray, y_pred: np.ndarray):  # -> list[dict[int, float]]:
    """
    Compute the class wise auc precision recall curve for a multiclass problem.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The aggregated auc precision recall curve class wise.
    """

    # TODO: quick fix done here.
    # if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0] and np.issubdtype(y_pred.dtype, np.floating)):
    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0]):
        raise ValueError(
            "y_true and y_pred need to be 2D with the same number of samples and dtype float, "
            "got {} and {} with dtype {} instead.".format(y_true.shape, y_pred.shape, y_pred.dtype)
        )

    num_classes = y_true.shape[1]
    y_true = y_true.astype(np.float)
    y_true = multiclass_assignments_to_labels(y_true)
    y_pred_torch = torch.from_numpy(y_pred)
    y_true_torch = torch.from_numpy(y_true)
    pytorch_pr_curve_precision, pytorch_pr_curve_recall, pytorch_pr_curve_thres = torchmetrics.functional.precision_recall_curve(
        y_pred_torch,
        y_true_torch,
        num_classes=num_classes)

    agg_test_pytorch_auc_pr_curve_class_wise = []
    np_agg_test_pytorch_auc_pr_curve_class_wise = np.full(len(pytorch_pr_curve_precision), np.nan)
    pytorch_auc_pr_curve_class_wise = dict()
    for ten in range(len(pytorch_pr_curve_precision)):
        val = torchmetrics.functional.auc(pytorch_pr_curve_recall[ten], pytorch_pr_curve_precision[ten]).numpy().item()
        pytorch_auc_pr_curve_class_wise[ten] = val
        np_agg_test_pytorch_auc_pr_curve_class_wise[ten] = val
    agg_test_pytorch_auc_pr_curve_class_wise.append(pytorch_auc_pr_curve_class_wise)

    return np_agg_test_pytorch_auc_pr_curve_class_wise

# TODO float
def multiclass_weighted_pytorch_auc_roc(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Compute the pytorch auc roc for a multiclass problem with average `weighted`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The auc roc score.
    """

    # TODO: quick fix done here.
    # if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0] and np.issubdtype(y_pred.dtype, np.floating)):
    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0]):
        raise ValueError(
            "y_true and y_pred needs to be 2D with same number of samples and dtype float, "
            "got {} and {} with dtype {} instead.".format(y_true.shape, y_pred.shape, y_pred.dtype)
        )

    num_classes = y_true.shape[1]
    y_true = y_true.astype(np.float)
    y_true = multiclass_assignments_to_labels(y_true)
    y_pred = y_pred.astype(np.float)

    y_pred_torch = torch.from_numpy(y_pred)
    y_true_torch = torch.from_numpy(y_true)
    return torchmetrics.functional.auroc(y_pred_torch,
                                         y_true_torch,
                                         num_classes=num_classes,
                                         average='weighted').numpy().item()

# TODO float
def multiclass_pytorch_auc_roc(y_true: np.ndarray, y_pred: np.ndarray): # -> list[float]:
    """
    Compute the pytorch auc roc for a multiclass problem with average `None`.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The auc roc score.
    """

    # TODO: quick fix done here.
    # if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0] and np.issubdtype(y_pred.dtype, np.floating)):
    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0]):
        raise ValueError(
            "y_true and y_pred needs to be 2D with the same number of samples and dtype float, "
            "got {} and {} with dtype {} instead.".format(y_true.shape, y_pred.shape, y_pred.dtype)
        )
    num_classes = y_true.shape[1]
    y_true = y_true.astype(np.float)
    y_true = multiclass_assignments_to_labels(y_true)
    y_pred = y_pred.astype(np.float)

    y_pred_torch = torch.from_numpy(y_pred)
    y_true_torch = torch.from_numpy(y_true)
    pytorch_auc_roc_class_wise = torchmetrics.functional.auroc(y_pred_torch,
                                                               y_true_torch,
                                                               num_classes=num_classes,
                                                               average=None)
    return [val.numpy().item() for val in pytorch_auc_roc_class_wise]


def multi_label_ranking_avg_precision_score(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Compute the label ranking based average precision score for a multi-label problem.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The precision score.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return label_ranking_average_precision_score(y_true=y_true, y_score=y_pred)


def multi_label_ranking_loss(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Compute the label ranking loss for a multi-label problem.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The precision score.
    :return: The loss.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return label_ranking_loss(y_true=y_true, y_score=y_pred)


def multi_label_normalized_discounted_cumulative_gain(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Compute the normalized discounted cumulative gain for a multi-label problem.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The gain.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return ndcg_score(y_true=y_true, y_score=y_pred)


def multiclass_top_1_accuracy(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Compute the top-1 accuracy for a multiclass problem.
    :param y_true: `numpy.array` of shape `(n_samples,)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The accuracy score.
    """

    if not (y_true.ndim == 1 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0]):
        raise ValueError(
            "y_true needs to be 1D and y_pred needs to be 2D with same number of samples, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return top_k_accuracy_score(y_true, y_pred, k=1, normalize=True, sample_weight=None,
                                labels=None)


def multiclass_top_3_accuracy(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Compute the top-3 accuracy for a multiclass problem.
    :param y_true: `numpy.array` of shape `(n_samples,)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The accuracy score.
    """

    if not (y_true.ndim == 1 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0]):
        raise ValueError(
            "y_true needs to be 1D and y_pred needs to be 2D with same number of samples, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return top_k_accuracy_score(y_true, y_pred, k=3, normalize=True, sample_weight=None,
                                labels=None)


def multiclass_top_5_accuracy(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Compute the top-5 accuracy for a multiclass problem.
    :param y_true: `numpy.array` of shape `(n_samples,)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape (n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The accuracy score.
    """

    if not (y_true.ndim == 1 and y_pred.ndim == 2 and y_true.shape[0] == y_pred.shape[0]):
        raise ValueError(
            "y_true needs to be 1D and y_pred needs to be 2D with same number of samples, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    top_5_acc = top_k_accuracy_score(y_true, y_pred, k=5, normalize=True, sample_weight=None, labels=None)
    return top_5_acc


def multiclass_log_loss(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    The logarithmic loss for a multiclass problem.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The loss.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return log_loss(y_true, y_pred)


def multi_label_log_loss(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    The logarithmic loss for a multi-label problem.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The loss.
    """

    if not (y_true.ndim == 2 and y_pred.ndim == 2 and y_true.shape == y_pred.shape):
        raise ValueError(
            "y_true and y_pred need to be 2D and have the same shape, "
            "got {} and {} instead.".format(y_true.shape, y_pred.shape)
        )

    return log_loss(y_true, y_pred)




def micro_precision(y_true, y_pred):
    """
    Calculate the micro precision, i.e. TP / (TP + FP).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The micro precision.
    """
    return precision_score(y_true, y_pred, average='micro')


def micro_recall(y_true, y_pred):
    """
    Calculate the micro recall, i.e.  TP / (TP + FN).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The micro recall.
    """
    return recall_score(y_true, y_pred, average='micro')


def micro_f1(y_true, y_pred):
    """
    Calculate the micro F1-score, i.e. 2 * (Precision * Recall) / (Precision + Recall).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The micro F1-score.
    """
    return f1_score(y_true, y_pred, average='micro')


def micro_f2(y_true, y_pred):
    """
    Calculate the micro F2-score (beta=2).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The micro F2-score.
    """
    return fbeta_score(y_true, y_pred, average='micro', beta=2)


def micro_jaccard(y_true, y_pred):
    """
    Calculate the micro Jaccard-score, i.e. TP / (TP + FP + FN).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The micro Jaccard-score.
    """
    return jaccard_score(y_true, y_pred, average='micro')


def macro_precision(y_true, y_pred):
    """
    Calculate the macro precision, i.e. TP / (TP + FP).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The macro precision.
    """
    return precision_score(y_true, y_pred, average='macro')


def macro_recall(y_true, y_pred):
    """
    Calculate the macro recall, i.e.  TP / (TP + FN).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The macro recall.
    """
    return recall_score(y_true, y_pred, average='macro')


def macro_f1(y_true, y_pred):
    """
    Calculate the macro F1-score, i.e. 2 * (Precision * Recall) / (Precision + Recall).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The macro F1-score.
    """
    return f1_score(y_true, y_pred, average='macro')


def weighted_f1(y_true, y_pred):
    """
    Calculate the macro F1-score, i.e. 2 * (Precision * Recall) / (Precision + Recall), weighted by the class support.

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The weighted macro F1-score.
    """
    return f1_score(y_true, y_pred, average='weighted')


def macro_f2(y_true, y_pred):
    """
    Calculate the macro F2-score (beta=2).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The macro F2-score.
    """
    return fbeta_score(y_true, y_pred, average='macro', beta=2)


def macro_jaccard(y_true, y_pred):
    """
    Calculate the macro Jaccard-score, i.e. TP / (TP + FP + FN).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: The macro Jaccard-score.
    """
    return jaccard_score(y_true, y_pred, average='macro')


def weighted_jaccard(y_true, y_pred):
    """
    Calculate the Jaccard-score for each label, and find their average, weighted by support, i. e., the number of true instances of each label instance.

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: The macro Jaccard-score.
    """
    return jaccard_score(y_true, y_pred, average="weighted")


def accuracy(y_true, y_pred):
    """
    Calculate the accuracy, i.e. (TP + TN) / (TP + FP + FN + TN).

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: Accuracy.
    """
    return accuracy_score(y_true, y_pred)


def error_rate(y_true, y_pred):
    """
    Calculate the error rate, i. e. error_rate = 1-accuracy
    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or class assignments.
    :return: Error Rate of typ `float`
    """
    return 1-accuracy_score(y_true, y_pred)


def balanced_multiclass_accuracy(y_true, y_pred):
    """
    Calculate the balanced accuracy, i.e. (Precision + Recall) / 2.

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: Accuracy.
    """
    if y_true.ndim > 1 or y_pred.ndim > 1:
        y_true = multiclass_assignments_to_labels(y_true)
        y_pred = multiclass_assignments_to_labels(y_pred)
    return balanced_accuracy_score(y_true, y_pred)


def mean_multilabel_confusion_matrix(y_true, y_pred):
    """
    Calculate the normalized mean confusion matrix across all classes.

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: `numpy.array` of shape `(n_classes, n_classes)`. Normalized mean confusion matrix.
    """
    cm_sum = np.sum(multilabel_confusion_matrix(y_true, y_pred, ), axis=0)
    return cm_sum / (len(y_pred) * np.max(cm_sum))


def mean_confidence(y_true, y_pred):
    """
    Calculate the mean confidence for continuous multiclass and multilabel classification outputs.

    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`. True class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples, n_classes)`. Predicted class assignments.

    :return: Mean confidence.
    """
    return 1 - np.sum(np.abs(y_true - y_pred)) / (y_true.shape[0] * y_true.shape[1])


def hamming(y_true, y_pred):
    """
    Calculate the average Hamming Loss.

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: Average Hamming Loss.
    """
    return hamming_loss(y_true, y_pred)


def log(y_true, y_pred):
    """
    Calculate the Logistic Loss.

    :param y_true: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. True labels or class assignments.
    :param y_pred: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Predicted labels or
            class assignments.
    :return: Logistic Loss.
    """
    return log_loss(y_true, y_pred)


def cohens_kappa(y1, y2, labels):
    """
    Calculate the Cohen's Kappa annotator agreement score according to :footcite:`cohen1960coefficient`.

    .. footbibliography::

    :param y1: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Labels or class assignments.
    :param y2: `numpy.array` of shape `(n_samples,)` or `(n_samples, n_classes)`. Labels or class assignments.
    :param labels: `list` of all possible labels.
    :return: Cohen's Kappa score.
    """
    cm = confusion_matrix(y1, y2, labels=labels)
    a = np.sum(np.diagonal(cm)) / np.sum(cm)
    e = 0
    for i in range(len(cm)):
        e += np.sum(cm[i, :]) * np.sum(cm[:, i]) / np.sum(cm) ** 2
    if e == 1:
        return 1.0  # case when y1 and y2 are equivalent in their annotation
    return (a - e) / (1 - e)


def pairwise_cohens_kappa(decision_tensor):
    """
    Calculate the average of pairwise Cohen's Kappa scores over all multiclass decision outputs.
    E.g., for 3 classifiers `(0,1,2)`, the agreement score is calculated for classifier tuples `(0,1)`, `(0,2)` and
    `(1,2)`. These scores are then averaged over all 3 classifiers.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of crisp multiclass decision outputs by different classifiers per sample.
    :return: Pairwise (averages) Cohen's Kappa score.
    """
    decision_tensor = np.array(decision_tensor)

    if determine_problem(decision_tensor) == Problem.MULTI_LABEL:
        decision_tensor = multilabel_to_multiclass_assignments(decision_tensor)

    n_classifiers = decision_tensor.shape[0]
    n_classes = decision_tensor.shape[2]
    indices = np.array(np.triu_indices(n_classifiers, k=1))
    sum_kappa = 0.0
    for i, j in zip(indices[0], indices[1]):
        decision_labels = multiclass_assignments_to_labels([decision_tensor[i], decision_tensor[j]])
        sum_kappa += cohens_kappa(decision_labels[0], decision_labels[1], labels=np.arange(n_classes))
    return sum_kappa / len(indices[0])


def __relations(y1, y2, y_true):
    """
    A helper function for calculating the correctness relations between two classifier outputs.
    `A` accumulates samples which are correctly classified by both classifiers, `B` accumulates those which are
    correctly classified by `c_1` but not by `c_2` and so on.
    """
    n_samples = len(y_true)
    a, b, c, d = 0, 0, 0, 0
    for i in range(n_samples):
        if np.all(y1[i] == y_true[i]) and np.all(y_true[i] == y2[i]):
            a += 1  # both classifiers are correct.
        elif np.all(y1[i] == y_true[i]) and np.any(y_true[i] != y2[i]):
            b += 1  # c1 is correct, c2 is wrong.
        elif np.any(y1[i] != y_true[i]) and np.all(y_true[i] == y2[i]):
            c += 1  # c1 is wrong, c2 is correct.
        elif np.any(y1[i] != y_true[i]) and np.any(y_true[i] != y2[i]):
            d += 1  # both classifiers are wrong.
    return a / n_samples, b / n_samples, c / n_samples, d / n_samples


def __pairwise_avg_score(decision_tensor, true_assignments, score_func, **kwargs):
    """
    A helper function for calculating pairwise average score statistics.
    """
    decision_tensor = np.array(decision_tensor)
    indices = np.array(np.triu_indices(decision_tensor.shape[0], k=1))
    scores = []
    for i, j in zip(indices[0], indices[1]):
        scores.append(score_func(decision_tensor[i], decision_tensor[j], true_assignments))

    if 'return_type' in kwargs:
        if kwargs['return_type'] == 'list':
            return indices, scores

    return np.mean(scores)


def correlation(y1, y2, y_true):
    """
    Calculate the correlation score for decision outputs of two classifiers according to Kuncheva
    :footcite:`kuncheva2014combining`.

    .. footbibliography::

    :param y1: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the first classifier.
    :param y2: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the second classifier.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Correlation score.
    """
    a, b, c, d = __relations(y1, y2, y_true)
    return (a * d - b * c) / np.sqrt((a + b) * (c + d) * (a + c) * (b + d))


def q_statistic(y1, y2, y_true):
    """
    Calculate the Q statistic score for decision outputs of two classifiers according to Yule
    :footcite:`udny1900association`.

    .. footbibliography::

    :param y1: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the first classifier.
    :param y2: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the second classifier.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Correlation score.
    """
    a, b, c, d = __relations(y1, y2, y_true)
    return (a * d - b * c) / (a * d + b * c)


def kappa_statistic(y1, y2, y_true):
    """
    Calculate the kappa score for decision outputs of two classifiers according to Kuncheva
    :footcite:`kuncheva2014combining`.

    .. footbibliography::

    :param y1: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the first classifier.
    :param y2: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the second classifier.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Kappa score.
    """
    a, b, c, d = __relations(y1, y2, y_true)
    return (2 * (a * d - b * c)) / ((a + b) * (b + d) + (a + c) * (c + d))


def disagreement(y1, y2, y_true):
    """
    Calculate the disagreement for decision outputs of two classifiers, i.e. the percentage of samples which are
    correctly classified by exactly one of the classifiers.

    :param y1: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the first classifier.
    :param y2: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the second classifier.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Disagreement score.
    """
    a, b, c, d = __relations(y1, y2, y_true)
    return b + c


def double_fault(y1, y2, y_true):
    """
    Calculate the double fault for decision outputs of two classifiers, i.e. the percentage of samples which are
    misclassified by both classifiers.

    :param y1: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the first classifier.
    :param y2: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the second classifier.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Double fault score.
    """
    a, b, c, d = __relations(y1, y2, y_true)
    return d


def abs_correlation(y1, y2, y_true):
    """
    Calculate the absolute correlation score for decision outputs of two classifiers.

    :param y1: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the first classifier.
    :param y2: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the second classifier.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Correlation score.
    """
    a, b, c, d = __relations(y1, y2, y_true)
    return np.abs((a * d - b * c) / np.sqrt((a + b) * (c + d) * (a + c) * (b + d)))


def abs_q_statistic(y1, y2, y_true):
    """
    Calculate the absolute Q statistic score for decision outputs of two classifiers.

    :param y1: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the first classifier.
    :param y2: `numpy.array` of shape `(n_samples, n_classes)`.
            Crisp multiclass decision outputs by the second classifier.
    :param y_true: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Correlation score.
    """
    a, b, c, d = __relations(y1, y2, y_true)
    return np.abs((a * d - b * c) / (a * d + b * c))


def pairwise_correlation(decision_tensor, true_assignments, **kwargs):
    """
    Calculate the average of the pairwise absolute correlation scores over all decision outputs.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of crisp multiclass decision outputs by different classifiers per sample.
    :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Pairwise correlation score.
    """
    if determine_problem(decision_tensor) == Problem.MULTI_LABEL:
        decision_tensor = multilabel_to_multiclass_assignments(decision_tensor)
        true_assignments = multilabel_to_multiclass_assignments(true_assignments)
    return __pairwise_avg_score(decision_tensor, true_assignments, abs_correlation, **kwargs)


def pairwise_q_statistic(decision_tensor, true_assignments):
    """
    Calculate the average of the pairwise absolute Q-statistic scores over all decision outputs.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of crisp multiclass decision outputs by different classifiers per sample.
    :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Pairwise correlation score.
    """
    if determine_problem(decision_tensor) == Problem.MULTI_LABEL:
        decision_tensor = multilabel_to_multiclass_assignments(decision_tensor)
        true_assignments = multilabel_to_multiclass_assignments(true_assignments)
    return __pairwise_avg_score(decision_tensor, true_assignments, abs_q_statistic)


def pairwise_kappa_statistic(decision_tensor, true_assignments, **kwargs):
    """
    Calculate the average of pairwise Kappa scores over all decision outputs.
    Multilabel class assignments are transformed to equivalent multiclass class assignments.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of crisp multiclass decision outputs by different classifiers per sample.
    :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Pairwise kappa score.
    """
    if determine_problem(decision_tensor) == Problem.MULTI_LABEL:
        decision_tensor = multilabel_to_multiclass_assignments(decision_tensor)
        true_assignments = multilabel_to_multiclass_assignments(true_assignments)
    return __pairwise_avg_score(decision_tensor, true_assignments, kappa_statistic, **kwargs)


def pairwise_disagreement(decision_tensor, true_assignments):
    """
    Calculate the average of pairwise disagreement scores over all decision outputs.
    Multilabel class assignments are transformed to equivalent multiclass class assignments.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of crisp multiclass decision outputs by different classifiers per sample.
    :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Pairwise disagreement score.
    """
    if determine_problem(decision_tensor) == Problem.MULTI_LABEL:
        decision_tensor = multilabel_to_multiclass_assignments(decision_tensor)
        true_assignments = multilabel_to_multiclass_assignments(true_assignments)
    return __pairwise_avg_score(decision_tensor, true_assignments, disagreement)


def pairwise_double_fault(decision_tensor, true_assignments, **kwargs):
    """
    Calculate the average of pairwise double fault scores over all decision outputs.
    Multilabel class assignments are transformed to equivalent multiclass class assignments.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of crisp multiclass decision outputs by different classifiers per sample.
    :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
            Matrix of crisp class assignments which are considered as true.
    :return: Pairwise double fault score.
    """
    if determine_problem(decision_tensor) == Problem.MULTI_LABEL:
        decision_tensor = multilabel_to_multiclass_assignments(decision_tensor)
        true_assignments = multilabel_to_multiclass_assignments(true_assignments)
    return __pairwise_avg_score(decision_tensor, true_assignments, double_fault, **kwargs)


def pairwise_euclidean_distance(decision_tensor):
    """
    Calculate the average of pairwise euclidean distance between decision matrices for the given classifiers.

    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
            Tensor of crisp multiclass decision outputs by different classifiers per sample.
    :return: Pairwise euclidean distance.
    """
    decision_tensor = np.array(decision_tensor)
    indices = np.array(np.triu_indices(decision_tensor.shape[0], k=1))
    scores = []
    for i, j in zip(indices[0], indices[1]):
        scores.append(np.mean(np.linalg.norm(decision_tensor[i] - decision_tensor[j], axis=1)))
    return np.mean(scores)




from pusion.core.combiner import UtilityBasedCombiner
from pusion.util.transformer import *
from pusion.util.constants import *


class BordaCountCombiner(UtilityBasedCombiner):
    """
    The :class:`BordaCountCombiner` (BC) is a decision fusion method that establishes a ranking between label
    assignments for a sample. This ranking is implicitly given by continuous support outputs and is mapped to different
    amounts of votes (:math:`0` of :math:`L` votes for the lowest support, and :math:`L-1` votes for the highest one).
    A class with the highest sum of these votes (borda counts) across all classifiers is considered as a winner for the
    final decision.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT)
    ]

    SHORT_NAME = 'BC'

    def __init__(self):
        UtilityBasedCombiner.__init__(self)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by the Borda Count (BC) method. Firstly, the continuous classification is mapped to a
        ranking with respect to available classes for each sample. Those rankings are then summed up across all
        classifiers to establish total votes (borda counts) for each class in a sample. The class with the highest
        number of borda counts is considered as decision fusion. Only continuous classification outputs are supported.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of continuous decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments which represents fused decisions.
                Axis 0 represents samples and axis 1 the class labels which are aligned with axis 2 in
                ``decision_tensor`` input tensor.
        """
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        fused_decisions = np.zeros_like(decision_tensor[0], dtype=int)

        for i in range(len(decision_profiles)):
            dp = decision_profiles[i]
            sort_indices = np.argsort(dp, axis=1)
            bc_dp = self.swap_index_and_values(sort_indices)
            fused_decisions[i, np.argmax(np.sum(bc_dp, axis=0))] = 1
        return fused_decisions

    def swap_index_and_values(self, m):
        s = np.zeros_like(m)
        for i in range(len(m)):
            for j in range(len(m[i])):
                s[i, m[i, j]] = j
        return s


class CRBordaCountCombiner(BordaCountCombiner):
    """
    The :class:`CRBordaCountCombiner` is a modification of :class:`BordaCountCombiner` that also supports
    complementary-redundant decision outputs. Therefore the input is transformed, such that all missing classification
    assignments are considered as `0`, respectively. To call :meth:`combine` a coverage needs to be set first
    by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT)
    ]

    def __init__(self):
        super().__init__()

    def combine(self, decision_outputs):
        """
        Combine complementary-redundant decision outputs by the Borda Count (BC) method. Firstly, the continuous
        classification is mapped to a ranking with respect to available classes for each sample. Those rankings are then
        summed up across all classifiers to establish total votes (borda counts) for each class in a sample. The class
        with the highest number of borda counts is considered as decision fusion. Only continuous classification outputs
        are supported.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains continuous decision outputs
                per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments which represents fused decisions.
                Axis 0 represents samples and axis 1 the class labels which are aligned with axis 2 in
                ``decision_tensor`` input tensor.

        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        return super().combine(t_decision_outputs)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.zeros((n_classifiers, n_decisions, n_classes))
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs



import numpy as np

from pusion.core.combiner import UtilityBasedCombiner
from pusion.util.constants import *


class MicroMajorityVoteCombiner(UtilityBasedCombiner):
    """
    The :class:`MicroMajorityVoteCombiner` (MIMV) is based on a variation of the general majority vote method.
    The fusion consists of a decision vector which results from the majority of assignments for each individual class.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT),
    ]

    SHORT_NAME = 'MIMV'

    def __init__(self):
        UtilityBasedCombiner.__init__(self)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by MIMV across all classifiers per class (micro).
        Only crisp classification outputs are supported.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of crisp decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments obtained by MIMV. Axis 0 represents samples and
                axis 1 the class labels which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        decision_tensor = decision_tensor - .5
        decision_sum = np.sum(decision_tensor, axis=0)
        fused_decisions = (decision_sum > 0) * np.ones_like(decision_sum)  # or (decision_sum >= 0)?
        return fused_decisions


class CRMicroMajorityVoteCombiner(MicroMajorityVoteCombiner):
    """
    The :class:`CRMicroMajorityVoteCombiner` is a modification of :class:`MicroMajorityVoteCombiner` that
    also supports complementary-redundant decision outputs. Therefore the input is transformed, such that all missing
    classification assignments are considered as a constant, respectively. To call :meth:`combine` a coverage needs to
    be set first by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
    ]

    def __init__(self):
        super().__init__()

    def combine(self, decision_outputs):
        """
        Combine decision outputs by MIMV across all classifiers per class (micro).
        Only crisp classification outputs are supported.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains crisp decision outputs
                per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments which are obtained by MIMV. Axis 0 represents
                samples and axis 1 all the class labels which are provided by the coverage.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        return super().combine(t_decision_outputs)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.full((n_classifiers, n_decisions, n_classes), .5)
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs



from pusion.auto.auto_combiner import AutoCombiner
from pusion.auto.generic_combiner import GenericCombiner
from pusion.core.combiner import Combiner, EvidenceBasedCombiner, TrainableCombiner
from pusion.model.configuration import Configuration


class DecisionProcessor:
    """
    :class:`DecisionProcessor` is the main user interface of the decision fusion framework. It provides all methods
    for selecting combiners including the :ref:`AutoCombiner <ac-cref>` and the :ref:`GenericCombiner <gc-cref>`.
    It also ensures uniformity and correct use of all `pusion.core` combiners.

    :param config: :class:`pusion.model.configuration.Configuration`. User-defined configuration.
    """
    def __init__(self, config: Configuration):
        self.config = config
        self.combiner = Combiner.obtain(config)
        self.evaluation = None

    def set_coverage(self, coverage):
        """
        Set the coverage in case of complementary-redundant classification data.

        :param coverage: `list` of `list` elements. Each inner list contains classes as integers covered by a
                classifier, which is identified by the positional index of the respective list. E.g., with
                ``[[0,1], [0,2,3]]`` the classes 0,1 are covered by the first classifier and 0,2,3 are covered by the
                second one.
        """
        self.combiner.set_coverage(coverage)

    def set_evidence(self, evidence):
        """
        Set the evidence for evidence-based combiners. The evidence is given by confusion matrices calculated
        according to Kuncheva :footcite:`kuncheva2014combining`.

        .. footbibliography::

        :param evidence: `list` of `numpy.array` elements of shape `(n_classes, n_classes)`. Confusion matrices
                for each ensemble classifier.
        """
        if isinstance(self.combiner, EvidenceBasedCombiner):
            self.combiner.set_evidence(evidence)

    def set_data_split_ratio(self, validation_size):
        """
        Set the size of the validation data used by the AutoCombiner to evaluate all applicable fusion methods in order
        to select the combiner with the best classification performance.
        Accordingly, the other data of size `1-validation_size` is used to train all individual combiners.

        :param validation_size: A `float` between `0` and `1.0`. Ratio of the validation data set.
        """
        if isinstance(self.combiner, AutoCombiner):
            self.combiner.set_validation_size(validation_size)
        raise TypeError("No AutoCombiner configuration.")

    def train(self, y_ensemble_valid, y_valid, **kwargs):
        """
        Train the combiner model determined by the configuration.

        .. warning::

            A trainable combiner is always trained with the validation dataset provided by ensemble classifiers.

        :param y_ensemble_valid: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.

                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param y_valid: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.

        :param **kwargs: The `**kwargs` parameter may be used to use additional test data for the AutoFusion selection
                procedure.
        """
        if isinstance(self.combiner, TrainableCombiner):
            if isinstance(self.combiner, AutoCombiner):
                self.combiner.train(y_ensemble_valid, y_valid, **kwargs)
            else:
                self.combiner.train(y_ensemble_valid, y_valid)

    def combine(self, y_ensemble_test):
        """
        Combine decision outputs using the combiner model determined by the configuration.

        :param y_ensemble_test: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
                `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
                due to the coverage.

                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: `numpy.array` of shape `(n_samples, n_classes)`. A matrix of crisp or continuous class assignments
                which represents fused decisions. Axis 0 represents samples and axis 1 the class labels which are
                aligned with axis 2 in ``y_ensemble_test`` input tensor.
        """
        return self.combiner.combine(y_ensemble_test)

    def get_multi_combiner_decision_output(self):
        """
        Retrieve the decision fusion outputs obtained by multiple combiners. This function is only callable for
        configurations including `AutoCombiner` or `GenericCombiner` as a method.

        :return: `list` of `numpy.array` elements of shape `(n_samples, n_classes)`.
                Fusion results obtained by multiple fusion methods.
                The list is aligned with the list of preselected fusion methods (retrievable by ``get_combiners()``).
        """
        if isinstance(self.combiner, GenericCombiner) or isinstance(self.combiner, AutoCombiner):
            return self.combiner.get_multi_combiner_decision_tensor()
        raise TypeError("No multi combiner output. Use combine(...) to retrieve the output of a single combiner.")

    def get_optimal_combiner(self, eval_metric=None):
        """
        Retrieve the combiner with the best classification performance obtained by the framework, i.e. the
        `AutoCombiner` or the `GenericCombiner`.
        In case of combining with the `GenericCombiner`, an `Evaluation` needs to be set by ``set_evaluation``.

        :return: The combiner object.
        """
        if type(self.combiner) is GenericCombiner and self.evaluation is None:
            raise TypeError("No evaluation set for determining the optimal method using GenericCombiner.")
        if type(self.combiner) is GenericCombiner:
            return self.evaluation.get_top_n_instances(n=1, metric=eval_metric)[0][0]
        elif type(self.combiner) is AutoCombiner:
            return self.combiner.get_selected_combiner()
        else:
            return self.combiner

    def get_combiners(self):
        """
        Retrieve combiners (core methods) which are preselected by the framework according to the auto-detected
        configuration.
        :return: `list` of combiner objects obtained by the `GenericCombiner` or `AutoCombiner`.
        """
        if isinstance(self.combiner, GenericCombiner) or isinstance(self.combiner, AutoCombiner):
            return self.combiner.get_combiners()
        raise TypeError("get_combiners() is not callable for a core combiner.")

    def get_combiner(self):
        """
        :return: Selected combiner object.
        """
        return self.combiner


    # def get_evaluation_results(self):
    #     """
    #     getter method to return the evaluation results of the combiners contained in the decision processor.
    #     :return: dict containing the evaluation reulsts of the combiners
    #     """
    #     if isinstance(self.combiner, GenericCombiner) or isinstance(self.combiner, AutoCombiner):
    #         problem, assignment_type, coverage_type = self.combiner.get_pac()
    #         combiner_type_selection = self.combiner.get_combiner_type_selection()
    #         optimal_comb = self.get_optimal_combiner()
    #
    #         report_dict = {
    #             'Problem': problem,
    #             'Assignment type': assignment_type,
    #             'Coverage type': coverage_type,
    #             'Combiner type selection': ', '.join([ct.__name__ for ct in combiner_type_selection]),
    #             'Compatible combiners': ', '.join([type(comb).__name__ for comb in self.get_combiners()]),
    #             'Optimal combiner': type(optimal_comb).__name__,
    #             'Classification performance': self.evaluation.get_performance_matrix(),
    #             'Evaluation metrics': self.evaluation.get_metrics(),
    #             'Instances': self.evaluation.get_instances()
    #         }
    #         return report_dict
    #     raise TypeError("get_evaluation_results() is not callable.")



    def report(self, eval_metric=None):
        """
        :return: The textual evaluation report.
        """
        if isinstance(self.combiner, GenericCombiner) or isinstance(self.combiner, AutoCombiner):
            problem, assignment_type, coverage_type = self.combiner.get_pac()
            combiner_type_selection = self.combiner.get_combiner_type_selection()
            optimal_comb = self.get_optimal_combiner(eval_metric=eval_metric)
            selection_metric = self.combiner.get_eval_metric().__name__ if isinstance(self.combiner, AutoCombiner) else (eval_metric.__name__ if eval_metric is not None else self.evaluation.metrics[0].__name__)

            report_dict = {
                'Problem': problem,
                'Assignment type': assignment_type,
                'Coverage type': coverage_type,
                'Combiner type selection': ', '.join([ct.__name__ for ct in combiner_type_selection]),
                'Compatible combiners': ', '.join([type(comb).__name__ for comb in self.get_combiners()]),
                'Selection metric': selection_metric,
                'Optimal combiner': type(optimal_comb).__name__,
                'Classification performance': '\n' + str(self.evaluation.get_report())
            }

            report_str = " " + type(self.combiner).__name__ + " - Report "
            report_str = report_str.center(90, '=') + "\n"
            tab_len = max([len(s) for s in report_dict.keys()])
            for k, v in report_dict.items():
                format_str = "{:>" + str(tab_len) + "}: {}\n"
                report_str += format_str.format(k, v)
            report_str += '=' * 90

            return report_str
        raise TypeError("report() is not callable for a core combiner.")

    def info(self):
        """
        Retrieve the information, the automatic combiner selection is based on.

        :return: `tuple` of the form `((A, B, C), D)`, whereby `A` represents the classification problem
                `('MULTI_CLASS' or 'MULTI_LABEL')`, `B` the assignment type `('CRISP' or 'CONTINUOUS')` and `C`
                the coverage type `('REDUNDANT', 'COMPLEMENTARY' or 'COMPLEMENTARY_REDUNDANT')`.

                `D` contains the combiner type selection as a `list`.
                Possible combiner types are `UtilityBasedCombiner`, `TrainableCombiner` and `EvidenceBasedCombiner`.
        """
        if isinstance(self.combiner, GenericCombiner) or isinstance(self.combiner, AutoCombiner):
            return self.combiner.get_pac(), self.combiner.get_combiner_type_selection()
        raise TypeError("info() is not callable for a core combiner.")

    def get_multi_combiner_runtimes(self):
        """
        Retrieve the train and combine runtime for each combiner used during a generic fusion.

        :return: A `tuple` of two lists of tuples describing the train and combine runtimes respectively.
                Each inner tuple key value indexes the list of preselected fusion methods
                (retrievable by ``get_combiners()``).
        """
        if isinstance(self.combiner, GenericCombiner) or isinstance(self.combiner, AutoCombiner):
            return self.combiner.get_multi_combiner_runtimes()
        raise TypeError("info_runtime() is not callable for a core combiner.")

    def set_evaluation(self, evaluation):
        """
        :param evaluation: :class:`pusion.control.evaluation.Evaluation` object, a combiner evaluation was
                performed with.
        """
        self.evaluation = evaluation

    def set_parallel(self, parallel=True):
        """
        Set whether the training and the combining of selected combiners should be executed sequentially or in parallel.
        :param parallel: If `True`, training and combining is performed in parallel respectively. Otherwise in sequence.
        """
        if isinstance(self.combiner, GenericCombiner) or isinstance(self.combiner, AutoCombiner):
            self.combiner.set_parallel(parallel)
        else:
            raise TypeError("set_parallel() is not callable for a core combiner.")



from pusion.core.decision_templates_combiner import *


class DempsterShaferCombiner(TrainableCombiner):
    """
    The :class:`DempsterShaferCombiner` (DS) fuses decision outputs by means of the Dempster Shafer evidence theory
    referenced by Polikar :footcite:`polikar2006ensemble` and Ghosh et al. :footcite:`ghosh2011evaluation`.
    DS involves computing the `proximity` and `belief` values per classifier and class, depending on a sample.
    Then, the total class support is calculated using the Dempster's rule as the product of belief values across all
    classifiers to each class, respectively. The class with the highest product is considered as a fused decision.
    DS shares the same training procedure with the :class:`DecisionTemplatesCombiner`.

    .. footbibliography::
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
    ]

    SHORT_NAME = 'DS'

    def __init__(self):
        TrainableCombiner.__init__(self)
        self.decision_templates = None
        self.distinct_labels = None

    def train(self, decision_tensor, true_assignments):
        """
        Train the Dempster Shafer Combiner model by precalculating decision templates from given decision outputs and
        true class assignments. Both continuous and crisp classification outputs are supported. This procedure involves
        calculations mean decision profiles (decision templates) for each true class assignment.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        dt_combiner = DecisionTemplatesCombiner()
        dt_combiner.train(decision_tensor, true_assignments)
        self.decision_templates = dt_combiner.get_decision_templates()
        self.distinct_labels = dt_combiner.get_distinct_labels()

    def combine(self, decision_tensor):
        """
        Combine decision outputs by using the Dempster Shafer method.
        Both continuous and crisp classification outputs are supported. Combining requires a trained
        :class:`DempsterShaferCombiner`.
        This procedure involves computing the proximity, the belief values, and the total class support using the
        Dempster's rule.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of either crisp or continuous class assignments which represents fused
                decisions obtained by the maximum class support. Axis 0 represents samples and axis 1 the class
                assignments which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        fused_decisions = np.zeros_like(decision_tensor[0])

        for i in range(len(decision_profiles)):
            dp = decision_profiles[i]
            n_label = len(self.decision_templates)
            n_classifiers = len(decision_tensor)
            
            # Compute proximity
            prox = np.empty((n_label, n_classifiers))  # Phi_{j,k}
            for j in range(n_label):
                dt = self.decision_templates[j]
                for k in range(n_classifiers):
                    d = 0.0
                    for j_ in range(n_label):
                        d = d + (1 + np.linalg.norm(self.decision_templates[j_][k] - dp[k]))**(-1)

                    prox[j, k] = (1 + np.linalg.norm(dt[k] - dp[k]))**(-1) / d

            # Compute belief
            bel = np.empty((n_label, n_classifiers))  # bel_{j,k}
            for j in range(n_label):
                for k in range(n_classifiers):
                    prod = 1.0
                    for j_ in range(n_label):
                        if j_ != j:
                            prod = prod * (1 - prox[j_, k])

                    bel[j, k] = prox[j, k] * prod / (1 - prox[j, k] * (1 - prod))

            # Compute support for each label (Dempster's rule)
            mu = np.zeros(n_label)
            for j in range(n_label):
                prod = 1.0
                for k in range(n_classifiers):
                    prod = prod * bel[j, k]
                mu[j] = prod

            # normalization
            mu = mu / np.sum(mu)
            fused_decisions[i] = self.distinct_labels[np.argmax(mu)]

        return fused_decisions


class CRDempsterShaferCombiner(DempsterShaferCombiner):
    """
    The :class:`CRDempsterShaferCombiner` is a modification of :class:`DempsterShaferCombiner` that
    also supports complementary-redundant decision outputs. Therefore the input is transformed, such that all missing
    classification assignments are considered as a constant, respectively. To use methods :meth:`train` and
    :meth:`combine` a coverage needs to be set first by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
    ]

    def __init__(self):
        super().__init__()

    def train(self, decision_outputs, true_assignments):
        """
        Train the Dempster Shafer Combiner model by precalculating decision templates from given decision outputs and
        true class assignments. Both continuous and crisp classification outputs are supported. This procedure involves
        calculations mean decision profiles (decision templates) for each true class assignment.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains either crisp or continuous
                decision outputs per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        super().train(t_decision_outputs, true_assignments)

    def combine(self, decision_outputs):
        """
        Combine decision outputs by using the Dempster Shafer method.
        Both continuous and crisp classification outputs are supported. Combining requires a trained
        :class:`DempsterShaferCombiner`.
        This procedure involves computing the proximity, the belief values, and the total class support using the
        Dempster's rule.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage. Each matrix corresponds to
                one of `n_classifiers` classifiers and contains crisp or continuous decision outputs per sample.

        :return: A matrix (`numpy.array`) of crisp or continuous class assignments which represents fused decisions.
                Axis 0 represents samples and axis 1 the class labels which are aligned with axis 2 in
                ``decision_tensor`` input tensor.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        return super().combine(t_decision_outputs)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.negative(np.ones((n_classifiers, n_decisions, n_classes)))
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs



from pandas import DataFrame


class Report:
    """
    :class:`Report` is a string representation of the performance matrix retrieved by :class:`Evaluation` methods.

    :param performance_matrix: `numpy.array` of shape `(n_instances, n_metrics)`. Performance matrix containing
            performance values for each set instance row-wise and each set performance metric column-wise.
    :param instances: `list` of instances been evaluated which is aligned with the performance_matrix on axis `0`.
    :param metrics: `list` of metric functions which is aligned with the performance_matrix on axis `1`.
    """
    def __init__(self, performance_matrix, instances, metrics):
        if performance_matrix.shape[0] != len(instances):
            raise TypeError("`performance_matrix` (axis 0) is not aligned with the number of instances.")
        if performance_matrix.shape[1] != len(metrics):
            raise TypeError("`performance_matrix` (axis 1) is not aligned with the number of metrics.")

        self.records = {}
        self.instance_names = self.__generate_instance_names(instances)
        self.metric_names = [metric if type(metric) is str else metric.__name__ for metric in metrics]

        for i, pv in enumerate(performance_matrix):
            self.records[self.instance_names[i]] = pv

    @staticmethod
    def __generate_instance_names(instances):
        instance_names = []
        for instance in instances:
            if isinstance(instance, str):
                instance_names.append(instance)
            else:
                instance_names.append(type(instance).__name__)
        instance_names_ = list(instance_names)
        for i, instance_name in enumerate(instance_names):
            if instance_names_.count(instance_name) > 1:
                instance_names[i] = "{:<35}".format(instance_name + " [" + str(i) + "]")
            else:
                instance_names[i] = "{:<35}".format(instance_name)
        return instance_names

    def __str__(self):
        return str(DataFrame.from_dict(data=self.records, orient='index', columns=self.metric_names))

    def get_data_frame(self):
        return DataFrame.from_dict(data=self.records, orient='index', columns=self.metric_names)



class ValidationError(Exception):
    def __init__(self, message):
        super(ValidationError, self).__init__(message)


class IllegalStateError(Exception):
    def __init__(self, message):
        super(IllegalStateError, self).__init__(message)



import numpy as np

from scipy import spatial

from pusion.core.combiner import UtilityBasedCombiner
from pusion.util.transformer import decision_tensor_to_decision_profiles
from pusion.util.constants import *


class CosineSimilarityCombiner(UtilityBasedCombiner):
    """
    The :class:`CosineSimilarityCombiner` considers the classification assignments to :math:`\\ell` classes as vectors
    from an :math:`\\ell`-dimensional vector space. The normalized cosine-similarity measure between two vectors
    :math:`x` and :math:`y` is calculated as

    .. math::
            cos(x,y) = \\dfrac{x\\cdot y}{|x||y|}\\ .

    The cosine-similarity is calculated pairwise and accumulated for each classifier for one specific sample.
    The fusion is represented by a classifier which shows the most similar classification output to the output of all
    competing classifiers.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.REDUNDANT)
    ]

    SHORT_NAME = 'COS'

    def __init__(self):
        UtilityBasedCombiner.__init__(self)

    def combine(self, decision_tensor):
        """
        Combine decision outputs with as an output that accommodates the highest cosine-similarity to the output of
        all competing classifiers. In other words, the best representative classification output among the others is
        selected according to the highest cumulative cosine-similarity. This method supports both, continuous and
        crisp classification outputs.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of either crisp or continuous decision outputs by different classifiers per sample.


        :return: A matrix (`numpy.array`) of either crisp or continuous class assignments which represents fused
                decisions obtained by the highest cumulative cosine-similarity. Axis 0 represents samples and axis 1 the
                class labels which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        fused_decisions = np.zeros_like(decision_tensor[0])
        decision_profiles = decision_tensor_to_decision_profiles(decision_tensor)
        for i in range(len(decision_profiles)):
            dp = decision_profiles[i]
            accumulated_cos_sim = np.zeros(len(dp))
            for j in range(len(dp)):
                for k in range(len(dp)):
                    if j != k and np.any(dp[j]) and np.any(dp[k]):
                        # Calculate the cosine distance (assumption: no zero elements)
                        accumulated_cos_sim[j] = accumulated_cos_sim[j] + (1 - spatial.distance.cosine(dp[j], dp[k]))
            fused_decisions[i] = dp[np.argmax(accumulated_cos_sim)]
        return fused_decisions


class CRCosineSimilarity(CosineSimilarityCombiner):
    """
    The :class:`CRCosineSimilarity` is a modification of :class:`CosineSimilarityCombiner` that also supports
    complementary-redundant decision outputs. Therefore the input is transformed, such that all missing classification
    assignments are considered as `0`, respectively. To call :meth:`combine` a coverage needs to be set first
    by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CONTINUOUS, CoverageType.COMPLEMENTARY_REDUNDANT),
        (Problem.MULTI_LABEL, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT)
    ]

    def __init__(self):
        super().__init__()

    def combine(self, decision_outputs):
        """
        Combine complementary-redundant decision outputs with as an output that accommodates the highest
        cosine-similarity to the output of all competing classifiers. In other words, the best representative
        classification output among the others is selected according to the highest cumulative cosine-similarity.
        This method supports both, continuous and crisp classification outputs.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage. Each matrix corresponds to
                one of `n_classifiers` classifiers and contains crisp or continuous decision outputs per sample.

        :return: A matrix (`numpy.array`) of crisp or continuous class assignments which represents fused decisions.
                Axis 0 represents samples and axis 1 the class labels which are aligned with axis 2 in
                ``decision_tensor`` input tensor.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        return super().combine(t_decision_outputs)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.zeros((n_classifiers, n_decisions, n_classes))
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs



class Problem:
    MULTI_LABEL = 'MULTI_LABEL'
    MULTI_CLASS = 'MULTI_CLASS'
    GENERIC = 'GENERIC'


class AssignmentType:
    CRISP = 'CRISP'
    CONTINUOUS = 'CONTINUOUS'
    GENERIC = 'GENERIC'


class CoverageType:
    REDUNDANT = 'REDUNDANT'
    COMPLEMENTARY = 'COMPLEMENTARY'
    COMPLEMENTARY_REDUNDANT = 'COMPLEMENTARY_REDUNDANT'
    GENERIC = 'GENERIC'


class EvidenceType:
    CONFUSION_MATRIX = 'CONFUSION_MATRIX'
    ACCURACY = 'ACCURACY'
    GENERIC = 'GENERIC'


class PAC:
    GENERIC = (Problem.GENERIC, AssignmentType.GENERIC, CoverageType.GENERIC)



import time


def p_train(index, combiner, decision_tensor, true_assignment, queue):
    """
    Encapsulates combiner's training procedure in order to be executable in a separate thread.

    :param index: Index that identifies the position within the thread safe queue.
    :param combiner: A combiner instance.
    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
            `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
            due to the coverage.
            Tensor of either crisp or continuous decision outputs by different classifiers per sample.
    :param true_assignment: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of either crisp or continuous class assignments which are considered true for each sample during
                the training procedure.
    :param queue: A thread safe queue for multiprocessing (e.g. `multiprocessing.managers.SyncManager.Queue`).
    """
    t_begin = time.perf_counter()
    combiner.train(decision_tensor, true_assignment)
    t_elapsed = time.perf_counter() - t_begin
    queue.put((index, combiner, t_elapsed))


def p_combine(index, combiner, decision_tensor, queue):
    """
    Encapsulates combiner's combine procedure in order to be executable in a separate thread.

    :param index: Index that identifies the position within the thread safe queue.
    :param combiner: A combiner instance.
    :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)` or a `list` of
            `numpy.array` elements of shape `(n_samples, n_classes')`, where `n_classes'` is classifier-specific
            due to the coverage.
            Tensor of either crisp or continuous decision outputs by different classifiers per sample.
    :param queue: A thread safe queue for multiprocessing (e.g. `multiprocessing.managers.SyncManager.Queue`).
    """
    t_begin = time.perf_counter()
    decision_matrix = combiner.combine(decision_tensor)
    t_elapsed = time.perf_counter() - t_begin
    queue.put((index, decision_matrix, t_elapsed))


def p_fit(index, classifier, x_train, y_train, queue):
    """
    Encapsulates classifier's fit procedure in order to be executable in a separate thread.

    :param index: Index that identifies the position within the thread safe queue.
    :param classifier: A classifier instance.
    :param x_train: Input data.
    :param y_train: Target values (i.e. labels or class assignments).
    :param queue: A thread safe queue for multiprocessing (e.g. `multiprocessing.managers.SyncManager.Queue`).
    """
    print("Train classifier: ", type(classifier).__name__, " [" + str(index) + "] ...")
    classifier.fit(x_train, y_train)
    queue.put((index, classifier))



from pusion.core.combiner import *
from pusion.util.constants import *
from pusion.util.transformer import *


class BehaviourKnowledgeSpaceCombiner(TrainableCombiner):
    """
    The :class:`BehaviourKnowledgeSpaceCombiner` (BKS) is adopted from the decision fusion method originally proposed by
    Huang, Suen et al. :footcite:`huang1993behavior`. BKS analyses the behaviour of multiple classifiers based on their
    classification outputs with respect to each available class.
    This behaviour is recorded by means of a lookup table, which is used for final combination of multiple
    classification outputs for a sample.

    .. footbibliography::

    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.REDUNDANT),
    ]

    SHORT_NAME = 'BKS'

    def __init__(self):
        super().__init__()
        self.unique_configs = None
        self.config_class_distribution = None
        self.n_classes = None

    def train(self, decision_tensor, true_assignments):
        """
        Train the Behaviour Knowledge Space model (BKS) by extracting the classification configuration from all
        classifiers and summarizing samples of each true class that leads to that configuration. This relationship is
        recorded in a lookup table. Only crisp classification outputs are supported.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of crisp decision outputs by different classifiers per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of crisp class assignments which are considered true for each sample during
                the training procedure.
        """
        configs = decision_tensor_to_configs(decision_tensor)
        unique_configs = np.unique(configs, axis=0)
        self.n_classes = np.shape(true_assignments)[1]
        n_unique_configs = np.shape(unique_configs)[0]
        config_class_distribution = np.zeros((n_unique_configs, self.n_classes), dtype=int)

        for i in range(n_unique_configs):
            unique_config = unique_configs[i]
            # Determine identical classification configurations for each of which
            # the number of samples is accumulated per true class assignment.
            b = np.array([np.all(unique_config == configs, axis=1)] * self.n_classes).transpose()
            config_class_distribution[i] = np.sum(true_assignments, axis=0, where=b)

        self.unique_configs = unique_configs
        self.config_class_distribution = np.array(config_class_distribution)

    def combine(self, decision_tensor):
        """
        Combine decision outputs by the Behaviour Knowledge Space (BKS) method. This procedure involves looking up the
        most representative class for a given classification output regarding the behaviour of all classifiers in the
        ensemble. Only crisp classification outputs are supported. If a trained lookup entry is not present for a
        certain classification configuration, no decision fusion can be made for the sample, which led to that
        configuration. In this case, the decision fusion is a zero vector.

        :param decision_tensor: `numpy.array` of shape `(n_classifiers, n_samples, n_classes)`.
                Tensor of crisp decision outputs by different classifiers per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments which are obtained by the best representative class
                for a certain classifier's behaviour per sample. Axis 0 represents samples and axis 1 the class labels
                which are aligned with axis 2 in ``decision_tensor`` input tensor.
        """
        configs = decision_tensor_to_configs(decision_tensor)
        fused_decisions = np.zeros((len(decision_tensor[0]), self.n_classes))

        for i in range(len(configs)):
            # perform a lookup in unique_configs
            lookup = np.where(np.all(configs[i] == self.unique_configs, axis=1))[0]
            if lookup.size > 0:
                uc_index = lookup[0]
                # set the class decision according to the maximum sample numbers for this config
                fused_decisions[i, self.config_class_distribution[uc_index].argmax()] = 1
        return fused_decisions


class CRBehaviourKnowledgeSpaceCombiner(BehaviourKnowledgeSpaceCombiner):
    """
    The :class:`CRBehaviourKnowledgeSpaceCombiner` is a modification of :class:`BehaviourKnowledgeSpaceCombiner` that
    also supports complementary-redundant decision outputs. Therefore the input is transformed, such that all missing
    classification assignments are considered as a constant, respectively. To use methods :meth:`train` and
    :meth:`combine` a coverage needs to be set first by the inherited :meth:`set_coverage` method.
    """

    _SUPPORTED_PAC = [
        (Problem.MULTI_CLASS, AssignmentType.CRISP, CoverageType.COMPLEMENTARY_REDUNDANT),
    ]

    def __init__(self):
        super().__init__()

    def train(self, decision_outputs, true_assignments):
        """
        Train the Behaviour Knowledge Space model (BKS) by extracting the classification configuration from all
        classifiers and summarizing samples of each true class that leads to that configuration. This relationship is
        recorded in a lookup table. Only crisp classification outputs are supported.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains crisp decision outputs
                per sample.

        :param true_assignments: `numpy.array` of shape `(n_samples, n_classes)`.
                Matrix of crisp class assignments which is considered true for each sample during
                the training procedure.
        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        super().train(t_decision_outputs, true_assignments)

    def combine(self, decision_outputs):
        """
        Combine decision outputs by the Behaviour Knowledge Space (BKS) method. This procedure involves looking up the
        most representative class for a given classification output regarding the behaviour of all classifiers in the
        ensemble. Only crisp classification outputs are supported. If a trained lookup entry is not present for a
        certain classification configuration, no decision fusion can be made for the sample, which led to that
        configuration. In this case, the decision fusion is a zero vector.

        :param decision_outputs: `list` of `numpy.array` matrices, each of shape `(n_samples, n_classes')`,
                where `n_classes'` is classifier-specific and described by the coverage.
                Each matrix corresponds to one of `n_classifiers` classifiers and contains crisp decision outputs
                per sample.

        :return: A matrix (`numpy.array`) of crisp class assignments which are obtained by the best representative class
                for a certain classifier's behaviour per sample. Axis 0 represents samples and axis 1 all the class
                labels which are provided by the coverage.

        """
        t_decision_outputs = self.__transform_to_uniform_decision_tensor(decision_outputs, self.coverage)
        return super().combine(t_decision_outputs)

    @staticmethod
    def __transform_to_uniform_decision_tensor(decision_outputs, coverage):
        n_classifiers = len(decision_outputs)
        n_decisions = len(decision_outputs[0])
        n_classes = len(np.unique(np.concatenate(coverage)))
        # tensor for transformed decision outputs
        t_decision_outputs = np.negative(np.ones((n_classifiers, n_decisions, n_classes)))
        for i in range(n_classifiers):
            t_decision_outputs[i, :, coverage[i]] = decision_outputs[i].T
        return t_decision_outputs
